import msgspec
from typing import List
import numpy as np


class NdArray(msgspec.Struct):
    data: bytearray  # Make it mutable.
    dtype: str
    shape: List[int]

    def to_numpy(self):
        """
        Attention!!! the internal buffer is mutable
        """
        return np.frombuffer(self.data, np.dtype(self.dtype)).reshape(self.shape)

    def to_view(self):
        return NdArrayView(memoryview(self.data), self.dtype, self.shape)


class NdArrayView(msgspec.Struct):
    data: memoryview
    dtype: str
    shape: List[int]

    @classmethod
    def from_numpy(cls, arr: np.ndarray):
        """
        Args:
            arr: numpy ndarray
            copy: copy the memory or keep it as a memoryview
        """
        return cls(arr.data, arr.dtype.str, arr.shape)

    def to_owned(self):
        return NdArray(bytearray(self.data), self.dtype, self.shape)



from tfrecord_slow import TfRecordWriter
from tqdm import tqdm
import numpy as np
from safetensors.numpy import save
import msgspec
from tfrecord_slow.msgpack import enc_hook
from .common import Message, NdArray


def bench(n: int = 1000):
    x = np.random.rand(1024, 1024)
    with TfRecordWriter.create("/tmp/test_writer.tfrec") as writer:
        for _ in tqdm(range(n)):
            writer.write(msgspec.msgpack.encode(Message(NdArray.from_numpy(x))))


bench()



from tfrecord_slow import TfRecordReader
from tqdm import tqdm
import numpy as np
import msgspec
from tfrecord_slow.msgpack import NdArray
from .common import Message


def bench(n: int = 1000):
    with TfRecordReader.open("/tmp/test_writer.tfrec") as reader:
        for data in tqdm(reader, total=n):
            msg = msgspec.msgpack.decode(data, type=Message)
            assert msg.x.to_numpy().shape == (1024, 1024)


bench()



from tfrecord_slow.utils.loader import RawTfrecordLoader, MsgpackTfrecordLoader
import pickle
import msgspec


def test_raw():
    assert pickle.dumps(RawTfrecordLoader(None))


class Example(msgspec.Struct):
    foo: int


def identity(x):
    return x


def test_msgpack():
    assert pickle.dumps(
        MsgpackTfrecordLoader(
            None,
            Example,
            identity,
        )
    )



import numpy as np
import msgspec
from tfrecord_slow.msgpack import NdArray, NdArrayView


def test_encode_decode():
    x = np.arange(100)
    arr = NdArrayView.from_numpy(x)
    buf = msgspec.msgpack.encode(arr)
    assert len(buf) > 0

    x1 = msgspec.msgpack.decode(buf, type=NdArray).to_numpy()
    assert x.dtype == x1.dtype
    assert np.allclose(x, x1)



import io
from typing import Optional
import struct
import logging

logger = logging.getLogger(__name__)


class TfRecordReader:
    def __init__(self, file: io.BufferedIOBase, check_integrity: bool = False) -> None:
        self._file = file
        self._check_integrity = check_integrity

        self._length_bytes = bytearray(8)
        self._crc_bytes = bytearray(4)
        self._data_bytes = bytearray(1024 * 1024)

    @classmethod
    def open(cls, path: str, check_integrity: bool = False):
        return cls(open(path, "rb"), check_integrity=check_integrity)

    def close(self):
        logger.debug("Close file: %s", self._file)
        self._file.close()

    def read(self) -> Optional[memoryview]:
        bytes_read = self._file.readinto(self._length_bytes)
        if bytes_read == 0:
            return
        elif bytes_read != 8:
            raise RuntimeError("Invalid tfrecord file: failed to read the record size.")

        if self._file.readinto(self._crc_bytes) != 4:
            raise RuntimeError("Invalid tfrecord file: failed to read the start token.")

        (length,) = struct.unpack("<Q", self._length_bytes)
        if length > len(self._data_bytes):
            self._data_bytes = self._data_bytes.zfill(length * 2)

        data_bytes_view = memoryview(self._data_bytes)[:length]
        if self._file.readinto(data_bytes_view) != length:
            raise RuntimeError("Invalid tfrecord file: failed to read the record.")
        if self._file.readinto(self._crc_bytes) != 4:
            raise RuntimeError("Invalid tfrecord file: failed to read the end token.")
        return data_bytes_view

    def __enter__(self):
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback):
        self.close()

    def __iter__(self):
        while True:
            data = self.read()
            if data is not None:
                yield data
            else:
                break

    def count(self):
        return sum(1 for _ in self)



import struct
from io import BufferedIOBase
from .utils.masked_crc import make_masked_crc


class TfRecordWriter:
    def __init__(self, file: BufferedIOBase) -> None:
        self._file = file

    @classmethod
    def create(cls, path: str, buffering: int = -1):
        return cls(open(path, mode="wb", buffering=buffering))

    def close(self):
        self._file.close()

    def flush(self):
        self._file.flush()

    def write(self, buf: bytes):
        length = len(buf)
        length_bytes = struct.pack("<Q", length)
        self._file.write(length_bytes)
        self._file.write(make_masked_crc(length_bytes))
        self._file.write(buf)
        self._file.write(make_masked_crc(buf))

    def __enter__(self):
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback):
        self.close()



from typing import TypeVar, Iterable, Iterator, Callable, Type
from io import BufferedIOBase
from tfrecord_slow.reader import TfRecordReader
import msgspec

T = TypeVar("T")


def _default_func(buf: memoryview):
    return buf.tobytes()


class RawTfrecordLoader:
    def __init__(
        self,
        datapipe: Iterable[BufferedIOBase],
        func: Callable[[memoryview], T] = _default_func,
        check_integrity: bool = False,
    ):
        self.datapipe = datapipe
        self.check_integrity = check_integrity
        self.func = func

    def __iter__(self) -> Iterator[T]:
        for fp in self.datapipe:
            reader = TfRecordReader(fp, check_integrity=self.check_integrity)
            for buf in reader:
                example = self.func(buf)
                yield example


S = TypeVar("S", bound=msgspec.Struct)


class MsgpackTfrecordLoader:
    def __init__(
        self,
        datapipe: Iterable[BufferedIOBase],
        spec: Type[S],
        check_integrity: bool = False,
    ):
        """
        Args:
            datapipe: iter of files
            spec: msgspec.Struct, must have owned memory
        """
        self.datapipe = datapipe
        self.check_integrity = check_integrity
        self.spec = spec

    def __iter__(self) -> Iterator[S]:
        decoder = msgspec.msgpack.Decoder(type=self.spec)
        for fp in self.datapipe:
            reader = TfRecordReader(fp, check_integrity=self.check_integrity)
            for buf in reader:
                example = decoder.decode(buf)
                yield example



"""
Make a new package for it.
"""

import argparse
import logging
from pathlib import Path
from typing import Optional
from .reader import TfRecordReader
from loguru import logger
from tqdm import tqdm


def get_args():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()
    add_count_args(subparsers.add_parser("count"))

    return parser.parse_args()


def add_count_args(parser: argparse.ArgumentParser):
    parser.set_defaults(func=count)
    parser.add_argument("path", type=Path, help="Path to tfrecord file or directory.")
    parser.add_argument("-m", "--mask", action="store_const", const="*.tfrec")
    parser.add_argument("-c", "--check", action="store_true", help="Check integrity.")


def count(args):
    path: Path = args.path
    mask: Optional[str] = args.mask
    check_integrity: bool = args.check

    if mask is not None:
        paths = list(path.glob(mask))
    else:
        paths = [path]

    for path in paths:
        with TfRecordReader.open(str(path), check_integrity=check_integrity) as reader:
            num_records = reader.count()
            print(f"file: {path}, records: {num_records}")


def main():
    args = get_args()
    print(args)

    if "func" in args:
        args.func(args)


if __name__ == "__main__":
    main()



import crc32c
import numpy as np
import struct

MASK = 0xA282EAD8
UINT32_MAX = 0xFFFFFFFF


def make_masked_crc(data: bytes) -> bytes:
    crc = np.uint32(crc32c.crc32c(data))
    masked = ((crc >> 15) | (crc << 17)) + MASK
    masked_bytes = struct.pack("<I", masked & UINT32_MAX)
    return masked_bytes


